/*
 * A minimal non-caching HTTP/HTTPS proxy.
 * 2002 by Thomas Knop <tknop@maxrelax.de>
 */
#include "config.h"

#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <strings.h>
#include <buffer.h>
#include <fmt.h>
#include <scan.h>
#include <netdb.h>
#include <netinet/in.h>
#ifdef USE_SYSLOG
#include <sys/syslog.h>
#endif

#include "sig.h"

int dropname = 0;
int loglevel = 1;
char *prog;
int ssltimeout = 5;
int timeout = 30;

struct request {
  int  ok;
  char buf[BUFFER_INSIZE];
  int  is_ssl;
  char *method;
  char *host;
  unsigned int port;
  char *path;
  char *proto;
} req;

buffer in;
char buf_in[BUFFER_INSIZE];
buffer out;
char buf_out[BUFFER_OUTSIZE];
char line[BUFFER_INSIZE];

static const char days[7][4] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
static const char month[12][4] = {"Jan","Feb","Mar","Apr","May","Jun",
                                  "Jul","Aug","Sep","Oct","Nov","Dez" };

#define buffer_puts2(b,s1,s2) buffer_puts3((b),(s1),(s2),0)

static void buffer_puts3(buffer *b, char *s1, char *s2, char *s3)
{
  if (s1) buffer_puts(b,s1);
  if (s2) buffer_puts(b,s2);
  if (s3) buffer_puts(b,s3);
}

static void log3(char *func, char *s1, char *s2, char *s3)
{
#ifdef USE_SYSLOG
  syslog(LOG_ERR, "%s%s%s %s %s", 
         func ? func : "", func ? ": " : "", 
         s1 ? s1 : "", s2 ? s2 : "", s3 ? s3 : "");
#else
  if (!dropname) buffer_puts2(buffer_2, prog, ": ");
  if (func) 
    buffer_puts2(buffer_2, func, ": ");
  buffer_puts3(buffer_2, s1, s2, s3);
  buffer_putsflush(buffer_2, "\n");
#endif
}

static void log_request(int code, long bytes)
{
  char *client;
#ifndef USE_SYSLOG
  char str[FMT_LONG];
#endif
  if (loglevel < 2) return;
  if(!(client = getenv("TCPREMOTEIP")))
    client = "";
#ifdef USE_SYSLOG
  syslog(LOG_INFO, "%d %s http%s://%s:%d/%s %ld bytes",
	 code, client, req.is_ssl ? "s" : "", 
	 req.host, req.port, req.path, bytes);
#else
  if (!dropname) buffer_puts2(buffer_2, prog, ": ");
  str[fmt_int(str, code)] = '\0';
  buffer_puts3(buffer_2, client, " ", str);
  str[fmt_long(str, bytes)] = '\0';
  buffer_puts3(buffer_2, " ", str, " ");
  buffer_puts3(buffer_2, req.method, " http", req.is_ssl ? "s" : "");
  str[fmt_uint(str, req.port)] = '\0';
  buffer_puts3(buffer_2, "://", req.host, ":");
  buffer_puts3(buffer_2, str, "/", req.path);
  buffer_putsflush(buffer_2, "\n");
#endif
}

/* implements strftime with format "%a, %d %b %Y %H:%M:%S GMT\r\n" */
static int fmt_rfc1123rn(char *str, struct tm *tm)
{
  register char *s = str;

  memcpy(s, days[tm->tm_wday], 3); s += 3;
  *s++ = ',';  
  *s++ = ' ';
  s += fmt_uint0(s, tm->tm_mday, 2);
  *s++ = ' ';
  memcpy(s, month[tm->tm_mon], 3); s += 3;
  *s++ = ' ';
  s += fmt_uint0(s, tm->tm_year+1900, 4);
  *s++ = ' ';
  s += fmt_uint0(s, tm->tm_hour, 2);
  *s++ = ':';
  s += fmt_uint0(s, tm->tm_min, 2);
  *s++ = ':';
  s += fmt_uint0(s, tm->tm_sec, 2);
  memcpy(s, " GMT", 4); s += 4;
  return (s-str);
}

static void http_error(unsigned int code, char* title, char* text)
{
  time_t now;
  char datestr[32];
  char codestr[FMT_ULONG];

  codestr[fmt_uint(codestr, code)] = '\0';
  if (req.ok)
    log_request(code, 0);
  if (loglevel > 0)
    log3(0, codestr, " ", title);
  now = time((time_t*) 0);
  datestr[fmt_rfc1123rn(datestr, gmtime(&now))] = '\0';
  buffer_puts3(buffer_1, "HTTP/1.0 ", codestr, " ");
  buffer_puts3(buffer_1, title, "\r\nServer: ", prog);
  buffer_puts3(buffer_1, "\r\nDate: ", datestr,  "\r\nContent-Type: text/html\r\n");
  buffer_puts3(buffer_1, "Connection:", " close\r\n\r\n", "<HTML><HEAD><TITLE>");
  buffer_puts3(buffer_1, codestr, " ", title);
  buffer_puts3(buffer_1, "</TITLE></HEAD>\n<BODY>\n<H2>", codestr, " ");
  buffer_puts3(buffer_1, title, "</H2>", text);
  buffer_puts(buffer_1, "\n<HR>\n</BODY></HTML>\n");
  buffer_flush(buffer_1);
  exit(1);
}

static int parse_request(void)
{
  register int i;
  register char *s;

  if ((i = buffer_getline(buffer_0, req.buf, sizeof(req.buf)-1)) <= 0)
    return -1;
  req.buf[i--] = '\0';
  while ((i > 0) && (req.buf[i] == '\n' || req.buf[i] == '\r'))
    req.buf[i--] = '\0';
#ifdef DEBUG
  if (loglevel == 2)
    log3("parse_request", "request = '", req.buf, "'");
#endif
  if(i < 0)
    return -1;
  /* get method */
  s = req.method = req.buf;
  while (*s != ' ')
    if(!*s++) return -2;
  *s++ = '\0'; 
#ifdef DEBUG
  if (loglevel == 2)
    log3("parse_request", "method = '", req.method, "'");
#endif
  if (!strncmp(s, "http://", 7)) { // http request
    req.is_ssl = 0;
    s += 7;
    /* get host */
    req.host = s;
    while (*s != ' ' && *s != ':' && *s != '/')
      if (!*s++) return -3;
    if (*s == ':') { /* port parameter */
      *s++ = '\0';
      s += scan_uint(s, &req.port);
    } else {
      *s++ = '\0';
      req.port = 80;
    }
#ifdef DEBUG
    if (loglevel == 2) {
      char str[FMT_ULONG];
      log3("parse_request", "(http) host = '", req.host, "'");
      str[fmt_ulong(str, (unsigned long)req.port)] = '\0';
      log3("parse_request", "(http) port = ", str, "");
    }
#endif
    /* get path */
    req.path = s;
    while (*s != ' ')
      if (!*s++) return -4;
    *s++ = '\0';
#ifdef DEBUG
    if (loglevel == 2)
      log3("parse_request", "(http) path = '", req.path, "'");
#endif
  } else if (!strcmp(req.method, "CONNECT")) { // ssl request
    req.is_ssl = 1;
    /* get host */
    req.host = s;
    while (*s != ' ' && *s != ':')
      if (!*s++) return -5;
    if (*s == ':') { /* port parameter */
      *s++ = '\0';
      s += scan_uint(s, &req.port);
    } else {
      *s++ = '\0';
      req.port = 443;
    }
    req.path = (char *)0;
#ifdef DEBUG
    if (loglevel == 2) {
      char str[FMT_ULONG];
      log3("parse_request", "(ssl) host = '", req.host, "'");
      str[fmt_ulong(str, (unsigned long)req.port)] = '\0';
      log3("parse_request", "(ssl) port =", str, "");
    }
#endif
    req.path = s-1;
    if (*s) s++;
  } else {
    return -6;
  }
  req.proto = s;
  while (*s != '\0')
    if(*s++ == ' ') return -7;
#ifdef DEBUG
  if (loglevel == 2)
    log3("parse_request", "proto = '", req.proto, "'");
#endif
  req.ok = 1;
  return 0;
}

static int open_client_socked(char *hostname, unsigned int port)
{
  char buf[4000];
  struct hostent h_ent, *h;
  int ht_err;
  struct sockaddr_in sa;
  int sockfd;

  memset(&sa, 0, sizeof sa);
  if (gethostbyname_r(hostname, &h_ent, buf, sizeof buf, &h, &ht_err))
    http_error(404, "Not Found", "Unknown host.");
  sa.sin_family = h->h_addrtype;
  memmove(&sa.sin_addr, h->h_addr, h->h_length);
  sa.sin_port = htons(port);
  sockfd = socket(sa.sin_family, SOCK_STREAM, 0);
  if (sockfd < 0)
    http_error(500, "Internal Error", "Couldn't create socket.");
  if (connect(sockfd, (struct sockaddr*) &sa, sizeof sa) < 0)
    http_error(503, "Service Unavailable", "Connection refused.");
  return sockfd;
}

static void sigalarm(int sig)
{
  http_error(408, "Request Timeout", "Request timed out.");
}

static void proxy_http(int sockfd)
{
  char c;
  int status = 200;
  long i, len = -1;

  buffer_init(&in, read, sockfd, buf_in, sizeof buf_in);
  buffer_init(&out, write, sockfd, buf_out, sizeof buf_out);
  /* forward client request to the server */
  buffer_puts3(&out, req.method, " /", req.path);
  buffer_puts3(&out, " ", req.proto, "\r\n");
  len = -1;
  while ((i = buffer_getline(buffer_0, line, sizeof line))) {
    do {
      line[i--] = '\0';
    } while (i > 0 && (line[i] == '\n' || line[i] == '\r'));
    if (i <= 0) 
      break;
    buffer_puts2(&out, line, "\r\n");
    alarm(timeout);
    /* keep track of content length */
    if (!strncasecmp(line, "Content-Length:", 15))
      scan_ulong(line + 16, &len);
  }
  buffer_puts(&out, "\r\n");
  /* forward possible content */
  for (;len > 0 && buffer_getc(buffer_0, &c); len--)
    buffer_put(&out, &c, 1);
  buffer_flush(&out);
  /* forward the server response to client */
  len = -2;
  alarm(timeout);
  while ((i = buffer_getline(&in, line, sizeof line))) {
    do {
      line[i--] = '\0';
    } while(i > 0 && (line[i] == '\n' || line[i] == '\r'));
    if (i <= 0) 
      break;
    if (len == -2) { /* first line, scan status */
      char *tmp = line;
      while (*tmp++ != ' ') ;
      scan_int(tmp, &status);
      len++;
    }
    if (!strncasecmp(line, "Content-Length:", 15))
      scan_long(line + 16, &len);
    if (strncasecmp(line, "Connection:", 11)) /* no keep alive */
      buffer_puts2(buffer_1, line, "\r\n");
    alarm(timeout);
  }
  buffer_puts2(buffer_1, "Connection:", " close\r\n\r\n");
  buffer_flush(buffer_1);
  if (len < 0) len = LONG_MAX;
  for (i = 0; (i < len) && buffer_getc(&in, &c); i++) {
    buffer_put(buffer_1, &c, 1);
    if (!(i % 8192))
      alarm(timeout);
  }
  buffer_flush(buffer_1);
  if (loglevel == 2) {
    log_request(status, i);
  }
}

static void proxy_ssl(int sockfd)
{
  struct timeval tv;
  fd_set fdset;
  long len = 0;
  register int n;

  /* Return SSL-proxy greeting header. */
  buffer_putsflush(buffer_1, "HTTP/1.0 200 Connection established\r\n\r\n");
  /* Now forward SSL packets in both directions until done. */
  tv.tv_sec = ssltimeout;
  tv.tv_usec = 0;
  for (;;) {
    FD_ZERO (&fdset);
    FD_SET (0, &fdset);
    FD_SET (sockfd, &fdset);
    if (select(sockfd + 1, &fdset, (fd_set*) 0, (fd_set*) 0, &tv) == 0)
      break;
    if (FD_ISSET (0, &fdset)) {
      if ((n = read(0, line, sizeof line)) <= 0)
	break;
      if ((n = write(sockfd, line, n)) <= 0)
	break;
    } else if (FD_ISSET(sockfd, &fdset)) {
      if ((n = read(sockfd, line, sizeof line)) <= 0)
	break;
      len += (unsigned long)n;
      if ((n = write(1, line, n) <= 0))
	break;
    }
  } 
  if (!len)
    http_error(408, "Request Timeout", "Request timed out.");
  else
    log_request(200, len);
}

static void doit()
{
  int sockfd, err;

  req.ok = 0;
  if ((err = parse_request()) != 0) {
    if (loglevel == 2) {
      char str[FMT_LONG];
      str[fmt_long(str, err)] = '\0';
      log3("parse_request", "status: ", str, 0);
    }
    http_error(400, "Bad Request", "Parse failed.");
  }
  sockfd = open_client_socked(req.host, req.port);
  if (req.is_ssl) {
    proxy_ssl(sockfd);
  } else {
    sig_set(SIGALRM, sigalarm);
    alarm(timeout);
    proxy_http(sockfd);
  }
  close(sockfd);
}

int main(int argc, char *argv[])
{
  int c;
  register char *s;

  for (s=prog=argv[0]; *s != '\0';)
    if (*s++ == '/') prog = s;
  while ((c=getopt(argc, argv, "dhqQvs:t:")) != -1) {
    switch (c) {
    case 'd': dropname = 1; break;
    case 't': scan_int(optarg, &timeout); break;
    case 's': scan_int(optarg, &ssltimeout); break;
    case 'q': loglevel = 0; break;
    case 'Q': loglevel = 1; break;
    case 'v': loglevel = 2; break;
    case 'h':
    default:
      buffer_puts3(buffer_2, 
		   "usage: ", prog, " [-dqQv] [-t timeout] [-s ssltimeout]\n");
      buffer_flush(buffer_2);
      exit(1);
    }
  }
#ifdef USE_SYSLOG
  openlog(prog, 0, LOG_DAEMON);
#endif

  doit();
  exit(0);
}
